function [dn ds dc dnc omega gamma lambda NCd] = EvolutionaryRateCount(s1,s2)


% Function to compute extended evolutionary rates with NG count method
% (c) S Pechmann (2011)


%default args
%gene_code = 1;
adjustStops = 0;


% initialize tables for codons, AAs, neighbours, codonusage, conservative
% mutations, syonymous mutations


codons = int2nt([ceil(1/16:1/16:4);repmat(ceil(1/4:1/4:4),1,4);repmat(1:4,1,16)])';
translation = nt2aa(codons,'GeneticCode',1,'AlternativeStartCodons',false);


% load conservative and synonymous mutations from file
conservative = load('conservative_mod20.dat');
synonymous = load('synonymous.dat');



% nine neighbours
v = eye(4,'uint8')==1;
u = ~v;
u = [u v v v; v u v v; v v u v; v v v u];
v = eye(16,'uint8')==1;
N = [u v v v; v u v v; v v u v; v v v u];

N(:,49) = 0;
N(49,:) = 0;
N(:,51) = 0;
N(51,:) = 0;
N(:,57) = 0;
N(57,:) = 0;


windowSize = inf;
slidingWindow = true;
adjustStops = true;


if isstruct(s1)
    s1 =seqfromstruct(s1);
end
if isstruct(s2)
    s2 = seqfromstruct(s2);
end

if(~(isnt(s1) && isnt(s2)))
    error('Bioinfo:dnds:IncorrectSequenceType',...
        'Inputs to DNDS must be DNA sequences.');
end

if ~isnumeric(s1)
    s1 = nt2int(s1);
end
if ~isnumeric(s2)
    s2 = nt2int(s2);
end

% use common shorter length multiple of 3
num_codons_original =  min(floor(numel(s1)/3),floor(numel(s2)/3));
num_nt = num_codons_original * 3;

s1 = reshape(s1(1:num_nt),3,num_codons_original);
s2 = reshape(s2(1:num_nt),3,num_codons_original);

if windowSize >= num_codons_original
    windowSize = num_codons_original;
    slidingWindow = false;
end



% gaps and ambiguous symbols are ignored, remove the whole codon
h = all(s1>=1,1) & all(s1<=4,1) & all(s2>=1,1) & all(s2<=4,1);

s1 = s1(:,h);
s2 = s2(:,h);
num_codons = sum(h);
num_nt = num_codons * 3;

seq1 = int2nt(s1);
seq2 = int2nt(s2);

% change nucleotide int triplets (1-4) to int codons (1-64)
s1 = (4.^(2:-1:0))*(double(s1)-1)+1;
s2 = (4.^(2:-1:0))*(double(s2)-1)+1;


%if adjustStops
    %remove also any stop codons from the computations
    g = translation(s1)~='*' & translation(s2)~='*'; 
    s1 = s1(g);
    s2 = s2(g);
    seq1 = seq1(:,g);
    seq2 = seq2(:,g);
    h(h)=g;
    num_codons = sum(h);
    num_nt = num_codons * 3;
%end

seq1 = nt2int(seq1);
seq2 = nt2int(seq2);


% Calculate the average potential mutations per class for every codon
% comparison in the sequences:
   
    
    % for each position: 6 parameters in 2 sequences -> average!
    
    
    allsynonymous = zeros(64,1);
    nonsynonymous = zeros(64,1);
    allconservative = zeros(64,1);
    nonconservative = zeros(64,1);
    
        
    
    for i=1:64
    
    allsynonymous(i,1) = length(find(N(i,:)==1 & synonymous(i,:)==1 )) / 3;
    nonsynonymous(i,1) = length(find(N(i,:)==1 & synonymous(i,:)==0 )) / 3;  
    allconservative(i,1) = length(find(N(i,:)==1 & synonymous(i,:)==0 & conservative(i,:)==1 )) / 3;
    nonconservative(i,1) = length(find(N(i,:)==1 & synonymous(i,:)==0 & conservative(i,:)==0 )) / 3;
    
    
    end
    
   
    
    
    
    z = sum(N(:,translation~='*'),2);   % number neighbours without stop codon
    
    if adjustStops == 1
        allsynonymous = allsynonymous * (num_codons*18/sum(z(s1)+z(s2)));
        nonsynonymous = nonsynonymous * (num_codons*18/sum(z(s1)+z(s2)));        
        allconservative = allconservative * (num_codons*18/sum(z(s1)+z(s2)));
        nonconservative = nonconservative * (num_codons*18/sum(z(s1)+z(s2)));

     
    end
    
    
    all_s = (allsynonymous(s1(:))     + allsynonymous(s2(:))) / 2;     % average number of potential differences
    all_n = (nonsynonymous(s1(:))   + nonsynonymous(s2(:))) / 2;
    all_c = (allconservative(s1(:))   + allconservative(s2(:))) / 2;
    all_nc = (nonconservative(s1(:)) + nonconservative(s2(:))) / 2;
   
    
    
   
    
    
    % Calculate the syn and nonsyn differences for every codon comparison
    % in the sequences:
   
    
    all_sd = zeros(num_codons,1);
    all_nd = zeros(num_codons,1);
    all_cd = zeros(num_codons,1);  
    all_ncd = zeros(num_codons,1);
    
    
   
    j = 1;
    for i = 1:3:num_nt-2
        
        
        codon1 = seq1(i:i+2);
        codon2 = seq2(i:i+2);
        
        %[codon1 codon2]
        
        %if (codon1 ~= codon2)
        [all_sd(j) all_nd(j) all_cd(j) all_ncd(j)] = pathways_diff(codon1,codon2, translation, synonymous, conservative);
        
        %end
        j = j + 1;
    end
    
    
    
    S = sum(all_s);
    N = sum(all_n);
    C = sum(all_c);
    NC = sum(all_nc);
    
    
    Sd = sum(all_sd);
    Nd = sum(all_nd);
    Cd = sum(all_cd);
    NCd = sum(all_ncd);
    
    
    
    

        if S==0, S = 1; end % rare case, it happens when Sd is also zero.
        if N==0, N = 1; end
        if C==0, C = 1; end
        if NC==0, NC = 1; end
              

    pS = (Sd ) ./ (S);  % Nei&Gojobori (1)
    pN = (Nd) ./ (N);
    pC = (Cd) ./ (C);
    pNC = (NCd ) ./ (NC);  % Nei&Gojobori (2)
    
    
   
    
    
    pS(pS >= 3/4 ) = NaN;
    pN(pN >= 3/4 ) = NaN;
    pC(pC >= 3/4 ) = NaN;
    pNC(pNC >= 3/4 ) = NaN;
    
    
  
    
    
    if all(~isnan(pS))
        ds = -(3/4) * log(1-((4/3)*pS));
    else
        ds = NaN;
    end
    
    if all(~isnan(pN))
        dn = -(3/4) * log(1-((4/3)*pN));
    else
        dn = NaN;
    end
    
    if all(~isnan(pC))
        dc = -(3/4) * log(1-((4/3)*pC));
    else
        dc = NaN;
    end
    
    if all(~isnan(pNC))
        dnc = -(3/4) * log(1-((4/3)*pNC));
    else
        dnc = NaN;
    end
    
    
    if ds ~=0
        omega = dn / ds;
    else
        omega = NaN;
    end
    
    
    if ds ~=0
        gamma = dc / ds;
    else
        gamma = NaN;
    end
    
    
    if dc ~=0
        lambda = dnc / dc;
    else
        lambda = NaN;
    end
    
    
    

    if all(isnan(pS))
        warning('Bioinfo:dnds:psLarge', ['Proportion of synonymous changes is greater than or equal to 3/4.\n' ...
            'Input sequences are too divergent, too short, or contain frame shifts.'])

   end


   end




    
    
%------------------------------------------
function [sd nd cd ncd] = pathways_diff(codon1, codon2, translation, synonymous, conservative)
% finds all possible paths from codon1 to codon2 and counts the number of
% synonymous changes and nonsynonymous changes. Note: no special case for
% nonsense mutations
nindx = find(codon1 ~= codon2);
sd = 0;
nd = 0;
cd = 0;
ncd = 0;



if(~isempty(nindx))
    order = perms(nindx);
    [paths changes] = size(order);
    first = repmat(codon1,paths,1);
    second = repmat(codon2,paths,1);
    A = first;
    for i = 1:changes -1
        ind1 = sub2ind(size(first), (1:numel(order(:,1)))', order(:,i));
        first(ind1) = second(ind1);
        A = [A first]; %#ok<AGROW>
    end
    A = [A second];
    
    Ar = reshape(A',1,numel(A));
    
    Acodons = zeros(length(Ar)/3,1); 
    
    p=1;
    for i = 1:3:((length(Ar))-2)
        tmp.codon = Ar(i:(i+2));
        
        Acodons(p,1) = (4.^(2:-1:0))*(double(tmp.codon)'-1)+1;
        p=p+1;
    end
  
 
    B  = reshape(Acodons,changes+1,paths)';
    if paths > 1 %adjustStops
        B = B(~any(translation(B)=='*',2),:);   % remove stopcodons from pathways!
        paths = size(B,1);
    end
    
    syn_p = 0;
    nsyn_p= 0;
    con_p = 0;
    non_p = 0;
        
    for j = 1:paths
        for i = 1:changes
         
            syn = synonymous(B(j,i), B(j,i+1));
            con = conservative(B(j,i), B(j,i+1));

    
    
            if syn == 1
                syn_p = syn_p + 1;             
                
            else
                nsyn_p = nsyn_p + 1;
                if con == 1
                    con_p = con_p + 1;
                else
                    non_p = non_p + 1;
                end
            end
        
        end 
    end
        
    


sd = syn_p / paths;
nd = nsyn_p / paths;
cd = con_p / paths;
ncd = non_p / paths;

end  

end
