function [rates params test like like2 NCsites Csites] = EvolutionaryRateML_lambda(s1,s2)
% 
% function to estimate evolutionary rates by maximum likelihood; 
% extension of the method by Yang and Goldman '94
%
% Sebastian Pechmann | Frydman Lab, Stanford | 2011
%


%default args
gene_code = 1;
%adjustStops = false;


% initialize tables for codons, AAs, neighbours, codonusage, conservative
% mutations, syonymous mutations

codon_table = geneticcode(gene_code);
codons = fieldnames(codon_table);
not_codon = (cellfun('length',codons) ~= 3);
codons(not_codon) = [];



% load conservative and synonymous mutations from file
%conservative = load('conservative_mod10.dat');

conservative = load('conservative_mod20.dat');
conservative([49 51 57],:) = [];
conservative(:,[49 51 57]) = [];

synonymous = load('synonymous.dat');
synonymous([49 51 57],:) = [];
synonymous(:,[49 51 57]) = [];

tstv = load('tstv.dat');
tstv([49 51 57],:) = [];
tstv(:,[49 51 57]) = [];




% nine neighbours
v = eye(4,'uint8')==1;
u = ~v;
u = [u v v v; v u v v; v v u v; v v v u];
v = eye(16,'uint8')==1;
N = [u v v v; v u v v; v v u v; v v v u];


length_readin = size(s1);
gap1 = sum(s1=='-') / length_readin(2);
gap2 = sum(s2=='-') / length_readin(2);



if isstruct(s1)
    s1 =seqfromstruct(s1);
end
if isstruct(s2)
    s2 = seqfromstruct(s2);
end

if(~(isnt(s1) && isnt(s2)))
    error('Bioinfo:dnds:IncorrectSequenceType',...
        'Inputs to DNDS must be DNA sequences.');
    return;
end

if ~isnumeric(s1)
    s1 = nt2int(s1);
end
if ~isnumeric(s2)
    s2 = nt2int(s2);
end

% use common shorter length multiple of 3
num_codons_original =  min(floor(numel(s1)/3),floor(numel(s2)/3));
num_nt = num_codons_original * 3;

s1 = reshape(s1(1:num_nt),3,num_codons_original);
s2 = reshape(s2(1:num_nt),3,num_codons_original);




% gaps and ambiguous symbols are ignored, remove the whole codon
h = all(s1>=1,1) & all(s1<=4,1) & all(s2>=1,1) & all(s2<=4,1);

s1 = s1(:,h);
s2 = s2(:,h);
num_codons = sum(h);
num_nt = num_codons * 3;

%seq1 = int2nt(s1);
%seq2 = int2nt(s2);
seq1 = int2nt(s1(:)');
seq2 = int2nt(s2(:)');


% change nucleotide int triplets (1-4) to int codons (1-64)
s1 = (4.^(2:-1:0))*(double(s1)-1)+1;
s2 = (4.^(2:-1:0))*(double(s2)-1)+1;


pi = get_pi(seq1, seq2);


sites = get_sites(s1,s2);%codons % get sites for likelihood calculation
sites([49 51 57],:) = [];
sites(:,[49 51 57]) = [];


N([49 51 57],:) = [];
N(:,[49 51 57]) = [];


sites2 = N .* sites;
num_all_sites = sum(sum(sites - diag(diag(sites)) ));
num_sites = sum(sum(sites2));

NCsites = sum(sum( sites2(conservative == 0) ));
Csites = sum(sum( sites2(conservative == 1))) - sum(diag(sites2)) ;

indQ  = build_indQ(synonymous, conservative, tstv);  %   stv: 1, sts: 2, nstv: 3, nsts: 4





% get initial estimates
[dn ds dc dnc omega gamma lambda NCd] = EvolutionaryRateCount(seq1,seq2);

% derive initial guess of omega, lambda etc.
[kappa, t] = HKY85_F84(seq1,seq2); % get initial guess at kappa and t



test=0;

options = optimset('fminsearch');

OptimOptions = optimset(options, 'Display','off');
%optimset('Display','off')

% 3 copute ML

% deal with edge cases
if ds==0 || isnan(ds) || dc==0 || isnan(dc) || dnc==0 || isnan(dnc) 
    if isinf(kappa) || kappa==0 %only estimate t by ML
        if dn==0 && ds==0 %equal sequences
            params = [0 0 0 0];
            like = 0;
            test=1;
        else    %estimate only t by ML
            [params like exitFlag] = fminsearch(@(x)get_likelihood(kappa,gamma, lambda,x(1),sites,pi,indQ),[kappa gamma lambda t],OptimOptions);
            params = [kappa gamma lambda params(1)];
            like = 0;
            test=2;
        end
    else        %estimate only kappa and t by ML
        [params like exitFlag] = fminsearch(@(x)get_likelihood(x(1),gamma, lambda,x(2),sites,pi,indQ),[kappa gamma lambda t],OptimOptions);
        params = [params(1) gamma lambda params(2)];
        test=3;
    end
elseif isinf(kappa) || kappa==0     %estimate all except kappa by ML
    [params like exitFlag] = fminsearch(@(x)get_likelihood(kappa,x(1),x(2),x(3),sites,pi,indQ),[kappa gamma lambda t],OptimOptions);
    params = [kappa params(1) params(2) params(3)];
    test=4;
elseif gap1 > 1/3 || gap2 > 1/3 %estimate only kappa and t by ML
        %params = [kappa omega lambda sigma t];
        [params like exitFlag] = fminsearch(@(x)get_likelihood(x(1),gamma, lambda,x(2),sites,pi,indQ),[kappa gamma lambda t],OptimOptions);
        params = [params(1) gamma lambda params(2)];
        test=6;
else                                %estimate all by ML
    [params like exitFlag] = fminsearch(@(x)get_likelihood(x(1),x(2),x(3),x(4),sites,pi,indQ),[kappa gamma lambda t],OptimOptions);
    test=5;
end
        


% 4 calculate rates

% Calculates dn ds

Q = build_Q(params(1), params(2), params(3),  pi, indQ);
rho = get_rho(indQ,Q,pi);
sumrho = sum(rho);


Qi = build_Q(params(1), 1, 1, pi, indQ);
rho_inf = get_rho(indQ,Qi,pi);
rates = params(4) * rho ./ (3 * rho_inf);

like = -like;



ll2 = logical( (indQ == 3) + (indQ == 4) + (indQ == 5) + (indQ == 6) );
P = expm(Q*t);                          
F = pi(ll2).*P(ll2);                               
Fn = max(0,F(sites(ll2)>0)); 
like2 = -sum((nonzeros(sites(ll2)).*log(Fn)));    %
like2 = - like2;
 




function indQ  = build_indQ(synonymous, conservative, tstv)
% Returns codon pair types corresponding to the matrix Q as follows:

indQ = zeros(61,61);    % 64x64
for i = 1:61
    for j = 1:61
        
        switch synonymous(i,j)   
            case 1 %synonymous substitution 
                
                indQ(i,j) = 1 + tstv(i,j);
                  
                            
            case 0 %nonsynonymous substitution      
                
              switch conservative(i,j) 
                  case 1    % conservative substitution
                      indQ(i,j) = 3 + tstv(i,j);                            
                      
                  case 0    % nonsynonymous substitution
                     indQ(i,j) = 5 + tstv(i,j);
                        
              end       
        end
    end
end
nan_locations = isnan(indQ);
indQ(nan_locations) = 0;      
        







function Q = build_Q(kappa, gamma, lambda,pi,Q)
%function Q = build_Q(kappa,omega, pi,Q)
% Returns the codon substitution rate matrix
if isinf(gamma)
    K = 0; gamma = 1;
else
    K = 1;
end
if isinf(kappa)
    J = 0; kappa = 1;
else
    J = 1;
end


syn_tv = (Q == 1);
syn_ts = (Q == 2);
con_tv = (Q == 3);
con_ts = (Q == 4);
nonsyn_tv = (Q == 5);
nonsyn_ts = (Q == 6);


Q(syn_tv) = pi(syn_tv);                       
Q(syn_ts) = kappa*pi(syn_ts);                 

Q(con_tv) = gamma*pi(con_tv);                 
Q(con_ts) = gamma*kappa*pi(con_ts);           
Q(nonsyn_tv) = lambda*gamma*pi(nonsyn_tv);                
Q(nonsyn_ts) = lambda*gamma*kappa*pi(nonsyn_ts);          


%%% for Q pi(j) is needed (the frequence of the codon mutated into!)

%Q(syn_tv) = pi(syn_tv);
%Q(syn_ts) = kappa*pi(syn_ts);
%Q(nsyn_tv) = omega*pi(nsyn_tv);
%Q(nsyn_ts) = omega*kappa*pi(nsyn_ts);



Q = Q + diag(-sum(Q,2));                 
D = -sum(diag(pi).*diag(Q));
if D~=0
    Q = Q ./ D ;        
else
    Q(:) = 0;
end




warning('off','MATLAB:log:logOfZero');
warning('off', 'MATLAB:illConditionedMatrix');


function like = get_likelihood(kappa,gamma,lambda,t,n,pi,Q)
%function like = get_likelihood(kappa,omega, t,n,pi,Q)

Q = build_Q(kappa,gamma,lambda,pi,Q);
P = expm(Q*t);                           
F = pi.*P;                               
Fn = max(0,F(n>0)); % removes small negative numbers that may appear due to 
                    % rounding errors, by setting these to 0, log(0)->-Inf 
                    % and like->Inf avoiding that FMINSEARCH picks this as
                    % a viable solution.  
like = -sum((nonzeros(n).*log(Fn)));    
   



function Q = build_Q2(kappa, gamma, lambda,pi,Q)
%function Q = build_Q(kappa,omega, pi,Q)
% Returns the codon substitution rate matrix
if isinf(gamma)
    K = 0; gamma = 1;
else
    K = 1;
end
if isinf(kappa)
    J = 0; kappa = 1;
else
    J = 1;
end


syn_tv = (Q == 1);
syn_ts = (Q == 2);
con_tv = (Q == 3);
con_ts = (Q == 4);
nonsyn_tv = (Q == 5);
nonsyn_ts = (Q == 6);


Q(syn_tv) = 0; %pi(syn_tv);                       
Q(syn_ts) = 0; %kappa*pi(syn_ts);                 

Q(con_tv) = 0; %gamma*pi(con_tv);                 
Q(con_ts) = 0; %gamma*kappa*pi(con_ts);           
Q(nonsyn_tv) = lambda*gamma*pi(nonsyn_tv);                
Q(nonsyn_ts) = lambda*gamma*kappa*pi(nonsyn_ts);          


%%% for Q pi(j) is needed (the frequence of the codon mutated into!)

%Q(syn_tv) = pi(syn_tv);
%Q(syn_ts) = kappa*pi(syn_ts);
%Q(nsyn_tv) = omega*pi(nsyn_tv);
%Q(nsyn_ts) = omega*kappa*pi(nsyn_ts);



Q = Q + diag(-sum(Q,2));                 
D = -sum(diag(pi).*diag(Q));
if D~=0
    Q = Q ./ D ;        
else
    Q(:) = 0;
end







function like = get_likelihood2(kappa,gamma,lambda,t,n,pi,Q)
%function like = get_likelihood(kappa,omega, t,n,pi,Q)
Q = build_Q2(kappa,gamma,lambda,pi,Q);
P = expm(Q*t);                           
F = pi.*P;                               
Fn = max(0,F(n>0)); % removes small negative numbers that may appear due to 
                    % rounding errors, by setting these to 0, log(0)->-Inf 
                    % and like->Inf avoiding that FMINSEARCH picks this as
                    % a viable solution.  
like = -sum((nonzeros(n).*log(Fn)));    



%function ci = get_CI(




%function [rhos rhon] = get_rho(Q,aa,pi)
function rho = get_rho(indQ,Q,pi)
% Returns the proportions of synonymous and nonsynonymous substitutions




syn_tv = (indQ == 1);
syn_ts = (indQ == 2);
con_tv = (indQ == 3);
con_ts = (indQ == 4);
nonsyn_tv = (indQ == 5);
nonsyn_ts = (indQ == 6);

%syn_tv = logical(syn_steady_tv + syn_unsteady_tv);
%syn_ts = logical(syn_steady_ts + syn_unsteady_ts);
%nsyn_tv = logical(con_steady_tv + con_unsteady_tv + nonsyn_steady_tv + nonsyn_unsteady_tv);
%nsyn_ts = logical(con_steady_ts + con_unsteady_ts + nonsyn_steady_ts + nonsyn_unsteady_ts);



%%% for rho, pi(i) is needed, the frequence of the codon that is mutated
%%% !!!

pi2 = pi';

%rho_syn = sum(pi2(syn_ts) .* Q(syn_ts)) + sum(pi2(syn_tv) .* Q(syn_tv));
%rho_nsyn = sum(pi2(nsyn_ts) .* Q(nsyn_ts)) + sum(pi2(nsyn_tv) .* Q(nsyn_tv));


rho_syn = sum(pi2(syn_ts) .* Q(syn_ts)) + sum(pi2(syn_tv) .* Q(syn_tv));

rho_con = sum(pi2(con_ts) .* Q(con_ts)) + sum(pi2(con_tv) .* Q(con_tv));
rho_nonsyn = sum(pi2(nonsyn_ts) .* Q(nonsyn_ts)) + sum(pi2(nonsyn_tv) .* Q(nonsyn_tv));





rho = [rho_syn rho_con rho_nonsyn];
%rho = [rho_nsyn rho_syn];


function sites = get_sites(s1, s2) %codons
% Returns a matrix with site counts for each codon pair ignoring gaps
%sites = zeros(length(codons), length(codons));
sites = zeros(64, 64);

len = min(length(s1), length(s2)); 
for i = 1:1:len
    sites(s1(i),s2(i)) = sites(s1(i),s2(i)) +1;
end
 



function [pi] = get_pi(seq1, seq2)

% Calculates equilibrium codon frequency via product of nucleotide
% frequency and adjust frequencies for terminator codons
[lastwmsg,lastwid]=lastwarn;
warnState=warning('off'); %#ok
seq = [seq1 seq2];
gene_code = 1;
[codon_freq cod_array] = codoncount(seq);
warning(warnState);
lastwarn(lastwmsg,lastwid);
nt_freq = ...
    [sum(sum(cod_array(1,:,:)))  sum(sum(cod_array(:,1,:))) sum(sum(cod_array(:,:,1))); ...
    sum(sum(cod_array(2,:,:)))  sum(sum(cod_array(:,2,:))) sum(sum(cod_array(:,:,2))); ...
    sum(sum(cod_array(3,:,:)))  sum(sum(cod_array(:,3,:))) sum(sum(cod_array(:,:,3))); ...
    sum(sum(cod_array(4,:,:)))  sum(sum(cod_array(:,4,:))) sum(sum(cod_array(:,:,4)))];
nt_freq  = nt_freq / sum(cod_array(:));
% get probs for termination codons
codon_table = geneticcode(gene_code);
codons = fieldnames(codon_table);
not_codon = (cellfun('length',codons) ~= 3);
codons(not_codon) = [];
translation = struct2cell(codon_table);
translation(not_codon) = [];
stpind = find(strcmp('*',translation));
stps = codons(stpind);
int_codon  = nt2int(cell2mat(codons));
int_stps  = nt2int(cell2mat(stps));
stp_total = sum(prod([nt_freq(int_stps(:,1),1) nt_freq(int_stps(:,2),2) nt_freq(int_stps(:,3),3)],2));
f = prod([nt_freq(int_codon(:,1),1) nt_freq(int_codon(:,2),2) nt_freq(int_codon(:,3),3)],2) ./ (1 - stp_total);
f(stpind) = [];
codons(stpind) = [];
int_codon(stpind,:) = [];

pi = repmat(f', length(f),1);



