function [K_HKY85, t] = HKY85_F84(s1,s2)
% Hasegawa distance taken from seqpdist modified according to Yang Nielsen
%
% Note: nucleotide frequencies are averaged between s1 and s2
%
% Let P and Q be the transition and transversional difference proportions
% and g the nucleotide frequencies over all the sequences:
%
% gr = g(1)+g(3);  % purines frequency
% gy = g(2)+g(4);  % pyrimidines frequency
% pr = g(1)*g(3);  % purines product
% py = g(2)*g(4);  % pyrimidines product
%
% Reference: Tamura and  Nei (1993) Molecular Biology and Evolution
% Reference: Yang and Nielsen(2000) Molecular Biology and Evolution

num_nt = min(length(s1), length(s2));


h = (s1(1:num_nt) ~='-') & (s2(1:num_nt)~='-');
map('ACGT- ') = 1:6;
temp1 = upper(char(s1(1:num_nt)));
temp2 = upper(char(s2(1:num_nt)));
g1 = accumarray(map(temp1(:))',1,[4 1]);
g2 = accumarray(map(temp2(:))',1,[4 1]);
g = (g1(1:4)+g2(1:4))/2;
g = g/ sum(g);
X  = accumarray(double([nt2int(s1(h))',nt2int(s2(h))']),1,[4 4]);

numPairs = sum(X(:));
numAGTransitions = sum(X([3 9]));
numCTTransitions = sum(X([8 14]));
numTransitions = numCTTransitions + numAGTransitions;
numTransversions = numPairs - sum(diag(X)) - numTransitions;
P = numTransitions / numPairs;
Q = numTransversions / numPairs;
gr = g(1)+g(3); % purines frequency
gy = g(2)+g(4); % pyrimidines frequency
pr = g(1)*g(3); % purines product
py = g(2)*g(4); % pyrimidines product

% Yang&Nielsen (8)
A = ( py+pr + (py*gr/gy+pr*gy/gr) * (1-(Q/(2*gy*gr))) - P/2 ) / (py/gy+pr/gr);
B = 1-Q/(2*gy*gr);

if B<=0
    warning('Bioinfo:dndsml:QLarge','Proportion of transversional nucleotide differences is too large.\nInput sequences are too divergent, too short, or contain frame shifts.')
    K_HKY85 = NaN;
    t = NaN;
    return
elseif A<=0
    warning('Bioinfo:dndsml:PLarge','Proportion of transitional nucleotide differences is too large.\nInput sequences are too divergent, too short, or contain frame shifts.')
    K_HKY85 = NaN;
    t = NaN;
    return
end

% Yang&Nielsen (9) (A and B must be positive !)
a = -log(A);
b = -log(B);

if Q==0 %(b==0)
    K_HKY85 = inf;
    t = 4*a*(py/gy+pr/gr); % taking the limit of Yang&Nielsen (10)
else
    K_F84 = a/b-1; % Yang&Nielsen (10)
    K_HKY85 = 1 + (((py/gy + pr/gr)*K_F84) / (py + pr));       % Yang&Nielsen (11)
    t = (4*py*(1+K_F84/gy) + 4*pr*(1+K_F84/gr) + 4*gy*gr) * b; % Yang&Nielsen (10)
end
K_HKY85 = max(0,K_HKY85);  % Kappa can't be negative, fixing rounding errors
t = max(0,t);              % t can't be negative, fixing rounding errors
