#!/usr/bin/python

"""
The MIT License (MIT)

Copyright (c) 2014 Sebastian Pechmann

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



import io, os, sys
import numpy as np
#from numpy import *
#from scipy.optimize import fmin
#from scipy.linalg import expm

from Bio import AlignIO
from evolutils import *


# Extended Evolutionary Rate Count model
# pechmann@stanford.edu



tables = np.load('tables.npz')
synonymous = tables['synonymous']
neighbours = tables['neighbours']
tstv = tables['tstv']

conservative = np.genfromtxt('conservative.dat')






codons = ['AAA', 'AAC', 'AAG', 'AAT', 'ACA', 'ACC', 'ACG', 'ACT', 'AGA', 'AGC', 'AGG', 'AGT', 'ATA', 'ATC', 'ATG', 'ATT', 'CAA', 'CAC', 'CAG', 'CAT', 'CCA', 'CCC', 'CCG', 'CCT', 'CGA', 'CGC', 'CGG', 'CGT', 'CTA', 'CTC', 'CTG', 'CTT', 'GAA', 'GAC', 'GAG', 'GAT', 'GCA', 'GCC', 'GCG', 'GCT', 'GGA', 'GGC', 'GGG', 'GGT', 'GTA', 'GTC', 'GTG', 'GTT', 'TAA', 'TAC', 'TAG', 'TAT', 'TCA', 'TCC', 'TCG', 'TCT', 'TGA', 'TGC', 'TGG', 'TGT', 'TTA', 'TTC', 'TTG', 'TTT']

stopcodons = ['TAA', 'TAG', 'TGA']
stopindx = [48, 50, 56]

neighbours[stopindx,:] = 0
neighbours[:,stopindx] = 0




def ExtendedEvolRateCount(s1, s2):


    num_codons_original = min( np.floor(len(s1)/float(3)), np.floor(len(s2)/float(3)) )
    num_nt = int(num_codons_original * 3)

    seq1 = []
    seq2 = []
    seqx1 = np.zeros((num_codons_original))
    seqx2 = np.zeros((num_codons_original))


    # Calculate the average potential mutations per class for every codon

    num_codons = 0
    for i in xrange(0,num_nt,3):
        current_codon1 = s1[i:i+3]
        current_codon2 = s2[i:i+3]

        if (current_codon1 in codons and current_codon1 not in stopcodons) and (current_codon2 in codons and current_codon2 not in stopcodons):
            seq1.append(current_codon1)
            seq2.append(current_codon2)

            seqx1[num_codons] = codons.index(current_codon1)
            seqx2[num_codons] = codons.index(current_codon2)
            num_codons += 1

    seqx1 = seqx1[0:num_codons].astype(int)
    seqx2 = seqx2[0:num_codons].astype(int)



    allsynonymous = np.zeros((64))
    allnonsynonymous = np.zeros((64))
    allconservative = np.zeros((64))
    allnonconservative = np.zeros((64))



    for i in range(64):

        current_neighbours = neighbours[:,i] == 1
        current_synonymous = synonymous[:,i] == 1
        current_nonsynonymous = synonymous[:,i] == 0
        current_conservative = conservative[:,i] == 1
        current_nonconservative = conservative[:,i] == 0 

        allsynonymous[i] = sum( current_neighbours * current_synonymous) / 3.
        allnonsynonymous[i] = sum( current_neighbours * current_nonsynonymous) / 3.
        allconservative[i] = sum( current_neighbours * current_nonsynonymous * current_conservative) / 3.
        allnonconservative[i] = sum( current_neighbours * current_nonsynonymous * current_nonconservative) / 3.


    # average over both sequences
    all_s =  (allsynonymous[seqx1] + allsynonymous[seqx2]) / 2.
    all_n =  (allnonsynonymous[seqx1] + allnonsynonymous[seqx2]) / 2.
    all_c =  (allconservative[seqx1] + allconservative[seqx2]) / 2.
    all_nc = (allnonconservative[seqx1] + allnonconservative[seqx2]) / 2.


    S   = sum(all_s)
    N   = sum(all_n)
    C   = sum(all_c)
    NC  = sum(all_nc)

    

    # Compute actual differences for each site
    all_sd  = np.zeros((num_codons))
    all_nd  = np.zeros((num_codons))
    all_cd  = np.zeros((num_codons))
    all_ncd = np.zeros((num_codons))

    # Compute the distribution between all possible 
    # mutational pathways between two codons, and their relative
    # contributions to the evolutionary rates

    count=0
    for i in range(num_codons):
        if (seq1[i] != seq2[i]):
            [all_sd[i], all_nd[i], all_cd[i], all_ncd[i]] = pathwaysdiff(seq1[i], seq2[i], codons, synonymous, conservative)
            count += 1 


    # Compute evolutionary rates after Nei & Gojobori
    Sd  = sum(all_sd)
    Nd  = sum(all_nd)
    Cd  = sum(all_cd)
    NCd = sum(all_ncd)

    if S == 0:
        S = 1

    if N == 0:
        N = 1

    if C == 0:
        C = 1

    if NC == 0:
        NC = 1

    pS  = Sd / float(S)
    pN  = Nd / float(N)
    pC  = Cd / float(C)
    pNC = NCd / float(NC)


    if (pS < (3./4.)):
        ds = -(3./4.) * math.log(1-((4./3.)*pS))
    else:
       ds = nan

    if (pN < (3./4.)):
        dn = -(3./4.) * math.log(1-((4./3.)*pN))
    else:
        dn = nan

    if (pC < (3./4.)):
        dc = -(3./4.) * math.log(1-((4./3.)*pC))
    else:
        dc = nan

    if (pNC < (3./4.)):
        dnc = -(3./4.) * math.log(1 -((4./3.)*pNC))
    else:
        dnc = nan


    if ds != 0:
        omega = dn/ds
    else: 
        omega = nan

    if ds != 0:
        gamma = dc/ds
    else: 
        gamma = nan

    if dc != 0:
        lmbda = dnc/dc
    else: 
        lmbda = nan


    #print "dS:", ds, "dN:", dn, "dC:", dc, "dNC:", dnc

    #print (Nd / float(N))
    #print (Cd+NCd) / float(C+NC)


    return ds, dn, dc, dnc, omega, gamma, lmbda






if __name__=='__main__':
    alnIN = sys.argv[1]
    alignment = AlignIO.read(open(alnIN), "fasta")
    ss1 = str( alignment[0].seq )
    ss2 = str( alignment[1].seq )

    ds, dn, dc, dnc, omega, gamma, lmbda =  ExtendedEvolRateCount(ss1, ss2 )
    print "dS:", round(ds,4), "dN:", round(dn, 4), "dC:", round(dc, 4), "dNC:", round(dnc,4), "omega:", round(omega, 4), "gamma:", round(gamma, 4), "lambda:", round(lmbda, 4)

