#!/usr/bin/python

"""
The MIT License (MIT)

Copyright (c) 2014 Sebastian Pechmann

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import io, os, sys
import numpy as np

from ExtendedEvolRateCount import ExtendedEvolRateCount
from evolutils import *
from Bio import AlignIO


# Extended Evolutionary Rate ML model
# 2014 | pechmann@stanford.edu


tables = np.load('tables.npz')
synonymous = tables['synonymous']
neighbours = tables['neighbours']
tstv = tables['tstv']

conservative = np.genfromtxt('conservative.dat')



codons = ['AAA', 'AAC', 'AAG', 'AAT', 'ACA', 'ACC', 'ACG', 'ACT', 'AGA', 'AGC', 'AGG', 'AGT', 'ATA', 'ATC', 'ATG', 'ATT', 'CAA', 'CAC', 'CAG', 'CAT', 'CCA', 'CCC', 'CCG', 'CCT', 'CGA', 'CGC', 'CGG', 'CGT', 'CTA', 'CTC', 'CTG', 'CTT', 'GAA', 'GAC', 'GAG', 'GAT', 'GCA', 'GCC', 'GCG', 'GCT', 'GGA', 'GGC', 'GGG', 'GGT', 'GTA', 'GTC', 'GTG', 'GTT', 'TAA', 'TAC', 'TAG', 'TAT', 'TCA', 'TCC', 'TCG', 'TCT', 'TGA', 'TGC', 'TGG', 'TGT', 'TTA', 'TTC', 'TTG', 'TTT']

stopcodons = ['TAA', 'TAG', 'TGA']
stopindx = [48, 50, 56]

neighbours[stopindx,:] = 0
neighbours[:,stopindx] = 0





def processSeqs(s1, s2):
    """
    input sequences as string
    returns for each sequence: vector of codons and vector of codon ids
    """

    num_codons_original = min( np.floor(len(s1)/float(3)), np.floor(len(s2)/float(3)) )
    num_nt = int(num_codons_original * 3)

    seq1 = []
    seq2 = []
    seqx1 = np.zeros((num_codons_original))
    seqx2 = np.zeros((num_codons_original))

    num_codons = 0
    for i in xrange(0,num_nt,3):
        current_codon1 = s1[i:i+3]
        current_codon2 = s2[i:i+3]

        if current_codon1 in codons and current_codon1 not in stopcodons and current_codon2 in codons and current_codon2 not in stopcodons:
            seq1.append(current_codon1)
            seq2.append(current_codon2)

            seqx1[num_codons] = codons.index(current_codon1)
            seqx2[num_codons] = codons.index(current_codon2)

        num_codons += 1

    seqx1 = seqx1[0:num_codons].astype(int)
    seqx2 = seqx2[0:num_codons].astype(int)


    return seq1, seq2, seqx1, seqx2



def optimize_parameters(kappa, gamma, lmbda, t, sites, Pi, indQ, ds, dc, dn, dnc):

    from scipy.optimize import fmin

    xtol_var = 0.0001
    maxiter_var = 100000
 
    if ds == 0 or np.isnan(ds) or dc == 0 or np.isnan(dc) or dnc == 0 or np.isnan(dnc):
        if np.isinf(kappa) or kappa == 0: 				# only estimate t by ML
            if dn == 0 and ds == 0:				# equal sequences
                opt_params = np.array([0, 0, 0, 0])
                like = 0
      
            else:						# estiate only t by ML
                param = (kappa, gamma, lmbda, sites, Pi, indQ)
                x = np.array([t])
                lfunc = lambda x, param: get_likelihood(param[0], param[1], param[2], x[0], param[3], param[4], param[5])
                xmin, like, iter, eval, warn = fmin(lfunc, x, args = (param,) , xtol=xtol_var, maxiter=maxiter_var, full_output = True, disp=0)
                opt_params = np.array([kappa, gamma, lmbda, xmin[0] ])

        else:							# estimate only kappa and t by ML
            param = (gamma, lmbda, sites, Pi, indQ)
            x = np.array([kappa, t])
            lfunc = lambda x, param: get_likelihood(x[0], param[0], param[1], x[1], param[2], param[3], param[4])
            xmin, like, iter, eval, warn = fmin(lfunc, x, args = (param,) , xtol=xtol_var, maxiter=maxiter_var, full_output = True, disp=0)
            opt_params = np.array([xmin[0], gamma, lmbda, xmin[1] ])        

    elif np.isinf(kappa) or kappa == 0:				# estimate all except kappa by ML
        param = (kappa, sites, Pi, indQ)
        x = np.array([gamma, lmbda, t])
        lfunc = lambda x, param: get_likelihood(param[0], x[0], x[1], x[2], param[1], param[2], param[3])
        xmin, like, iter, eval, warn = fmin(lfunc, x, args = (param,) , xtol=xtol_var, maxiter=maxiter_var, full_output = True, disp=0)
        opt_params = np.array([kappa, xmin[0], xmin[1], xmin[2] ])    


    #elif gap > 1/3
    #    print "sequences too divergent"


    else: 							# estimate all by ML
        param = (sites, Pi, indQ)
        x = np.array([kappa, gamma, lmbda, t])
        lfunc = lambda x, param: get_likelihood(x[0], x[1], x[2], x[3], param[0], param[1], param[2])
        opt_params, like, iter, eval, warn = fmin(lfunc, x, args = (param,) , xtol=0.0001, maxiter=100000, full_output = True, disp=0)


    return opt_params, -like





def calculate_rates(kappa, gamma, lmbda, t, Pi, indQ):

    Q = build_Q(kappa, gamma, lmbda, Pi, indQ)
    rho = get_rho(indQ, Q, Pi)
    #sumrho = sum(rho)

    Qi = build_Q(kappa, 1, 1, Pi, indQ)
    rho_inf = get_rho(indQ, Qi, Pi)
    rates = t * rho / (3. * rho_inf)

    return rates



def ExtendedEvolRateML(s1, s2):
    

    # get sequence vectors
    seq1, seq2, seqx1, seqx2 = processSeqs(s1, s2)

    # get codon frequencies
    Pi = get_Pi2(seq1, seq2, codons)
    Pi[stopindx,:] = 0
    Pi[:,stopindx] = 0

    # get sites
    sites = get_sites(seqx1, seqx2)
    sites[stopindx,:] = 0
    sites[:,stopindx] = 0

    # get index substitution matrix
    indQ = build_indQ(synonymous, conservative, tstv)

    # get initial estimates through count model and HKY85
    [ds, dn, dc, dnc, omega, gamma, lmbda] = ExtendedEvolRateCount(s1, s2) 
    [kappa, t] = HKY85_F84(s1, s2)

    # maximum likelihood parameter estimation
    opt_params, like = optimize_parameters(kappa, gamma, lmbda, t, sites, Pi, indQ, ds, dn, dc, dnc)

    # compute evolutionary rates
    rates = calculate_rates(opt_params[0], opt_params[1], opt_params[2], opt_params[3], Pi, indQ)


    return rates, opt_params, like


if __name__=='__main__':
    alnIN = sys.argv[1]
    alignment = AlignIO.read(open(alnIN), "fasta")
    s1 = str( alignment[0].seq )
    s2 = str( alignment[1].seq )

    rates, opt_params, like = ExtendedEvolRateML(s1, s2)

    print "dS:", round(rates[0],4), "dC:", round(rates[1], 4), "dNC:", round(rates[2],4), "kappa:", round(opt_params[0],4), "gamma:", round(opt_params[1], 4), "lambda:", round(opt_params[2], 4), "t:", round(opt_params[3],4), "like:", round(like, 4)

