"""
The MIT License (MIT)

Copyright (c) 2014 Sebastian Pechmann

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


""" Utility function to compute extended evolutionary rates. 


written by S Pechmann (pechmann@stanford.edu)


List of functions
~~~~~~~~~~~~~~~~~~

pathwaydiff: averages over all possible mutational pathways between 2 codons

HKY85_F84: computes Hasegawa distance, adapted from Yang & Nielsen (2000)

get_Pi2: computes codon frequences from input sequences

get_sites: gets site counts for each codon pair

build_indQ: returns codon pair types

build_Q: computes codon substitution matrix

get_rho: computes evolutionary rates

get_likelihood: computes likelihood

"""



import io, os, sys
import numpy as np
import math




def pathwaysdiff(codon1, codon2, codons, synonymous, conservative):
    """
    All possible pathways between two codons
    """

    import itertools
    sd, nd, cd, ncd = (0,)*4

    # Get all possible evolutionary paths between two codons
    codon1 = np.array(list(codon1))
    codon2 = np.array(list(codon2))
    c1 = ''.join(codon1)

    diffs = np.where(codon1 != codon2)                     # nt positiosn at which codons differ
    diffs = diffs[0].tolist()
    order = np.array(list( itertools.permutations(diffs) ) )  # all possible permutations
    (paths, changes) = order.shape
    pathways = np.zeros((paths, changes+1))
    pathways[:,0] = codons.index(c1)

    for i in range(paths):
        codon = np.copy(codon1)
        for j in range(changes):
            mutation = order[i,j]
            codon[mutation] = codon2[mutation]
            c =  ''.join(codon)
            ci = codons.index(c)
            pathways[i,j+1] = ci

    # Remove paths with stop codons
    if paths > 1:
        pathways = pathways[~(pathways==48).any(1), :]
        pathways = pathways[~(pathways==50).any(1), :]
        pathways = pathways[~(pathways==56).any(1), :]
    paths2 = pathways.shape[0]

    # Classify all mutations for all paths
    syn_p, nsyn_p, con_p, non_p = (float(0),)*4

    for i in range(paths2):
        for j in range(changes):
            first  = int(pathways[i,j])
            second = int(pathways[i,j+1])

            syn = synonymous[first, second]
            con = conservative[first, second]

            if (syn == 1):
                syn_p += 1
            else:
                nsyn_p += 1

                if (con == 1):
                    con_p += 1
                else:
                    non_p += 1

    dff = np.array([syn_p, nsyn_p, con_p, non_p])
    dff = dff /float(paths2)
    dff = np.around(dff, 5)
    return dff





def HKY85_F84(s1, s2):
    """
    Hasegawa distance taken from seqpdist modified according to Yang Nielsen

    Note: nucleotide frequencies are averaged between s1 and s2
 
    Let P and Q be the transition and transversional difference proportions
    and g the nucleotide frequencies over all the sequences:

    gr = g(1)+g(3);  % purines frequency
    gy = g(2)+g(4);  % pyrimidines frequency
    pr = g(1)*g(3);  % purines product
    py = g(2)*g(4);  % pyrimidines product

    Reference: Tamura and  Nei (1993) Molecular Biology and Evolution
    Reference: Yang and Nielsen(2000) Molecular Biology and Evolution
    """


    nucleotides = ['A', 'C', 'G', 'T']

    s1 = np.array(list(s1))
    s2 = np.array(list(s2))

    num_nt = min(len(s1), len(s2))

    g1 = np.array([sum(s1=='A'), sum(s1=='C'), sum(s1=='G'), sum(s1=='T')])
    g2 = np.array([sum(s2=='A'), sum(s2=='C'), sum(s2=='G'), sum(s2=='T')])
    g = (g1 + g2)/2.
    g /= sum(g)

    num_pairs = 0
    num_ts = 0
    num_tv = 0

    for i in xrange(0,num_nt):
        current_nt1 = s1[i]
        current_nt2 = s2[i]
        if current_nt1 in nucleotides and current_nt2 in nucleotides:
            num_pairs += 1
            if (current_nt1 == 'A' and current_nt2 == 'G'):
                num_ts +=1
            elif (current_nt1 == 'G' and current_nt2 == 'A'):
                num_ts +=1
            elif (current_nt1 == 'C' and current_nt2 == 'T'):
                num_ts +=1
            elif (current_nt1 == 'T' and current_nt2 == 'C'):
                num_ts +=1
  
            elif (current_nt1 != current_nt2): 
                num_tv +=1
 
    P = num_ts / float(num_pairs)
    Q = num_tv / float(num_pairs)

    gr = g[0]+g[2]; # purines frequency
    gy = g[1]+g[3]; # pyrimidines frequency
    pr = g[0]*g[2]; # purines product
    py = g[1]*g[3]; # pyrimidines product

    # Yang&Nielsen (8)
    A = (py+pr + (py*gr/gy + pr*gy/gr) * (1-(Q/(2*gy*gr))) - P/2.)  / (py/gy + pr/gr)
    B = 1 - (Q / (2*gy*gr) )

    if B <= 0:
        print 'Proportion of transversional nucleotide differences is too large.\nInput sequences are too divergent, too short, or contain frame shifts.'
        K_HKY85 = nan
        t = nan

    elif A <= 0:
        print 'Proportion of transitional nucleotide differences is too large.\nInput sequences are too divergent, too short, or contain frame shifts.'
        K_HKY85 = nan
        t = nan

    # Yang&Nielsen (9) (A and B must be positive !)
    a = -np.log(A)
    b = -np.log(B)

    if Q == 0:	#(b==0)
        K_HKY85 = np.inf
        t = 4*a*(py/gy + pr/gr) 	# taking the limit of Yang&Nielsen (10)
    else:
        K_F84 = a/float(b)-1
        K_HKY85 = 1 + (((py/gy + pr/gr)*K_F84) / (py + pr))
        t = (4*py*(1+K_F84/gy) + 4*pr*(1+K_F84/gr) + 4*gy*gr) * b

    K_HKY85 = max(0,K_HKY85) 	# Kappa can't be negative, fixing rounding errors
    t = max(0,t)             	# t can't be negative, fixing rounding errors

    return round(K_HKY85, 4), round(t, 4)







def get_Pi2(seq1, seq2, codons):

    stopcodons = ['TAA', 'TAG', 'TGA']
    nt = {'A':0, 'C':1, 'G':2, 'T':3}


    #get nt freqs

    if len(seq1) != len(seq2):
        print "sequences are of different length!"

    sequence =  "".join(seq1) + "".join(seq2)
    sequence = np.array( list(str(sequence) ) )

    number_codons = np.floor( len(sequence)/3. )
    ntarray = np.reshape(sequence, (number_codons,3))

    out = np.zeros(( 4,3 ))

    for i in range(3):
        out[0,i] = sum(ntarray[:,i] == 'A')
        out[1,i] = sum(ntarray[:,i] == 'C')
        out[2,i] = sum(ntarray[:,i] == 'G')
        out[3,i] = sum(ntarray[:,i] == 'T')
    out /= number_codons

    
    f = np.zeros(( 64 ))
    for i in range(64):
        current_codon = list(codons[i])
        f[i] = out[ int(nt[current_codon[0]]), 0] * out[ int(nt[current_codon[1]]), 1] * out[ int(nt[current_codon[2]]), 2]
 
    stp = 0.
    for i in stopcodons:
        current_codon = i
        stp += out[ int(nt[current_codon[0]]), 0] * out[ int(nt[current_codon[1]]), 1] * out[ int(nt[current_codon[2]]), 2]

    f /= (1-stp)

    Pi = np.kron(np.ones((64,1)), np.around(f, 14) )
    return Pi


   


def get_sites(sx1, sx2):
    sites = np.zeros((64, 64))
    seq_len = min( len(sx1), len(sx2) )
    for i in range(seq_len):
        sites[ sx1[i], sx2[i] ] += 1
    return sites




def build_indQ(synonymous, conservative, tstv):
    indQ = np.zeros((64,64))
    for i in range(64):
        for j in range(64):
            if synonymous[i,j] == 1:
                indQ[i,j] = 1 + tstv[i,j]
            else:
                if conservative[i,j] == 1:
                    indQ[i,j] = 3 + tstv[i,j]
                else:
                    indQ[i,j] = 5 + tstv[i,j]

    nan_locations = np.isnan(indQ)
    indQ[nan_locations] = 0

    stopindx = [48, 50, 56]
    indQ[stopindx,:] = 0
    indQ[:,stopindx] = 0
 
    return indQ




def build_Q(kappa, gamma, lmbda, Pi, indexQ):
    # Returns the codon substitution matrix
    Q = np.zeros((64, 64))

    if np.isinf(kappa):
        J = 0
        kappa = 1
    else:
        J = 1

    syn_tv = (indexQ == 1)
    syn_ts = (indexQ == 2)
    con_tv = (indexQ == 3)
    con_ts = (indexQ == 4)
    nonsyn_tv = (indexQ == 5)
    nonsyn_ts = (indexQ == 6)

    Q[syn_tv] = J*Pi[syn_tv]
    Q[syn_ts] = kappa*Pi[syn_ts]
    Q[con_tv] = J*gamma*Pi[con_tv]
    Q[con_ts] = gamma*kappa*Pi[con_ts]
    Q[nonsyn_tv] = J*lmbda*gamma*Pi[nonsyn_tv]
    Q[nonsyn_ts] = lmbda*gamma*kappa*Pi[nonsyn_ts]

    Q = Q + np.diag((-Q.sum(axis=1)))
    D = -sum(np.diag(Pi) * np.diag(Q) )  

    if D != 0:
        Q /= D                        # Yang&Nielsen, 2000
    else:
        Q[:,:] = 0

    return Q



def get_rho(indQ, Q, Pi):
    syn_tv = (indQ == 1)
    syn_ts = (indQ == 2)
    con_tv = (indQ == 3)
    con_ts = (indQ == 4)
    nonsyn_tv = (indQ == 5)
    nonsyn_ts = (indQ == 6)

    Pi2 = np.copy( Pi.transpose() )

    rho_syn = sum(Pi2[syn_tv + syn_ts]*Q[syn_tv + syn_ts]) 
    rho_con = sum(Pi2[con_tv + con_ts]*Q[con_tv + con_ts])
    rho_nonsym = sum(Pi2[nonsyn_tv + nonsyn_ts]*Q[nonsyn_tv + nonsyn_ts])
    rho = np.array([rho_syn, rho_con, rho_nonsym])

    return rho 




def get_likelihood(kappa, gamma, lmbda, t, n, Pi, Q):
    from scipy.linalg import expm
    np.seterr(divide='ignore')

    indexQ = np.copy(Q)
    
    Q = build_Q(kappa, gamma, lmbda, Pi, indexQ)
    P = expm(Q*t)
    F = Pi * P
    Fn = np.where(F[n>0] < 0, 0, F[n>0]) #max(0, F[n>0])     	
    nn = n[n!=0]						# nonzero sites
    logFn = np.log(Fn)						# patches log of 0
    Fn[Fn==-np.inf]=0
    like = -sum( nn * logFn )
 
    return like

